from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from betterconf.caster import AbstractCaster


class BetterconfError(Exception):
    pass


class ImpossibleToCastError(BetterconfError):
    def __init__(self, val: str, caster: "AbstractCaster"):
        self.val = val
        self.caster = caster
        self.message = f'Could not cast "{val}" with {caster.__class__.__name__}'
        super().__init__(self.message)


class VariableNotFoundError(BetterconfError):
    def __init__(self, variable_name: str):
        self.variable_name = variable_name
        self.message = f"Variable ({variable_name}) hasn't been found"
        super().__init__(self.message)



import dataclasses
import typing

_T = typing.TypeVar("_T")
_T2 = typing.TypeVar("_T2")


class _Special:
    pass


def is_special(tp: typing.Any) -> bool:
    return isinstance(tp, _Special)


@dataclasses.dataclass(eq=True, frozen=True)
class AliasSpecial(_Special):
    tp: type
    alias: str

    @classmethod
    def __class_getitem__(cls, *args: typing.Tuple[type, str]) -> typing.Self:
        parsed = args[0]

        return cls(parsed[0], parsed[1])


if typing.TYPE_CHECKING:
    Alias = typing.Annotated

else:
    Alias = AliasSpecial



import typing
from betterconf.provider import AbstractProvider
from betterconf._field import _NO_DEFAULT, _Field as Field  # type: ignore
from betterconf._specials import is_special, AliasSpecial
from betterconf.caster import BUILTIN_CASTERS
from dataclasses import dataclass
from betterconf.exceptions import BetterconfError

FT = typing.TypeVar("FT")


@dataclass
class Prefix:
    prefix: str
    delimiter: str = "_"

    @staticmethod
    def process_name(name: str, prefix: typing.Optional["Prefix"] = None) -> str:
        if not prefix:
            return name

        return f"{prefix.prefix}{prefix.delimiter}{name}"


class ConfigProto(typing.Protocol):
    __bc_subconfig__: typing.ClassVar[bool]
    __bc_inner__: typing.ClassVar["ConfigInner"]
    __bc_prefix__: typing.ClassVar[typing.Optional[Prefix]]
    __bc_provider__: typing.ClassVar[typing.Optional[AbstractProvider]]


@dataclass
class FieldInfo(typing.Generic[FT]):
    name_in_python: str
    field: Field[FT]

    @classmethod
    def parse_into(
        cls,
        src: type,
        name: str,
        annotation: FT,
        provider: typing.Optional[AbstractProvider] = None,
        prefix: typing.Optional[Prefix] = None,
    ) -> typing.Self:
        # for types like int, str, etc AND not initialized with `field`
        if annotation in BUILTIN_CASTERS and name not in src.__dict__:
            caster = BUILTIN_CASTERS[annotation]
            name_in_field = Prefix.process_name(name, prefix)
            return cls(
                name_in_python=name,
                field=Field(name=name_in_field, caster=caster, provider=provider),
            )

        # if it is our special annotations
        elif is_special(annotation):
            default = getattr(src, name, _NO_DEFAULT)

            if isinstance(annotation, AliasSpecial):
                name_in_field = Prefix.process_name(annotation.alias, prefix)

                field_info = cls.parse_into(
                    src, annotation.alias, annotation.tp, provider
                )
                field_info.name_in_python = name
                if isinstance(default, Field):
                    # var: Alias[str, "VAR"] = field(...)
                    field_info.field = default

                elif default is not _NO_DEFAULT:
                    field_info.field.default = default

                field_info.field.name = name_in_field

                if not field_info.field.provider:
                    field_info.field.provider = provider

                return field_info

        elif (
            annotation in BUILTIN_CASTERS
            and name in src.__dict__
            and isinstance(getattr(src, name), Field)
        ):
            field = typing.cast(Field[FT], getattr(src, name))
            field.caster = BUILTIN_CASTERS[annotation]

            if not field.provider:
                field.provider = provider
            if not field.name:
                field.name = name

            return cls(name_in_python=name, field=field)
        elif (
            annotation not in BUILTIN_CASTERS
            and name in src.__dict__
            and isinstance(getattr(src, name), Field)
        ):
            field: Field[FT] = getattr(src, name)
            if not field.provider:
                field.provider = provider
            if not field.name:
                field.name = name

            return cls(name_in_python=name, field=field)

        elif (
            annotation in BUILTIN_CASTERS
            and name in src.__dict__
            and isinstance(getattr(src, name), annotation)
        ):
            name_in_field = Prefix.process_name(name, prefix)
            val: typing.Any = getattr(src, name)
            field = Field(name=name_in_field, default=val, provider=provider)
            return cls(name_in_python=name, field=field)

        elif (
            annotation in BUILTIN_CASTERS
            and name in src.__dict__
            and not isinstance(getattr(src, name), annotation)
        ):
            raise BetterconfError(
                f"You try to set the value {repr(getattr(src, name))} for the field with name '{name}', that has type {annotation}.\nThe type {type(getattr(src, name))} is not assignable to type {annotation}"
            )

        raise BetterconfError(
            "Something bad happened.\nBetterconf can't deal with this kind of value.\nProbably you've tried to use something like 'dict' in a constant manner. For this special case use 'constant_field`"
        )


@dataclass
class SubConfigInfo:
    name: str
    cfg: typing.Type[ConfigProto]

    @classmethod
    def parse_into(
        cls,
        src: typing.Type[ConfigProto],
        provider: typing.Optional[AbstractProvider] = None,
        prefix: typing.Optional[Prefix] = None,
    ) -> typing.Self:
        if not src.__bc_provider__:
            src.__bc_provider__ = provider

        inner = ConfigInner.parse_into(src, provider, prefix)
        src.__bc_inner__ = inner
        return cls(name=src.__name__, cfg=src)


@dataclass
class ConfigInner:
    fields: typing.List[FieldInfo[typing.Any]]
    sub_configs: typing.List[SubConfigInfo]

    @classmethod
    def parse_into(
        cls,
        cfg: typing.Type[ConfigProto],
        provider: typing.Optional[AbstractProvider] = None,
        prefix: typing.Optional[Prefix] = None,
    ) -> typing.Self:
        try:
            annotations = typing.get_type_hints(cfg, include_extras=True)
        except TypeError:
            annotations: typing.Dict[str, typing.Any] = {}

        fields_info: typing.List[FieldInfo[typing.Any]] = []
        for name, annotation in annotations.items():
            if name in cfg.__dict__:
                name_in_python = Prefix.process_name(name, prefix) if prefix is not None else name
                field: Field[typing.Any] = cfg.__dict__[name]

                if not field.provider:
                    field.provider = provider

                fields_info.append(FieldInfo(name_in_python, field))
                continue

            parsed = FieldInfo[typing.Any].parse_into(
                cfg,
                name,
                annotation,
                provider,
                prefix,
            )
            fields_info.append(parsed)

        sub_configs: typing.List[SubConfigInfo] = []
        for name, element in cfg.__dict__.items():
            if getattr(element, "__bc_subconfig__", False):
                parsed = SubConfigInfo.parse_into(element, provider, prefix)
                sub_configs.append(parsed)
            elif isinstance(element, Field):
                if element.name is None:
                    element.name = (
                        name if prefix is None else Prefix.process_name(name, prefix)
                    )

                field_info = FieldInfo(name, typing.cast("Field[typing.Any]", element))
                fields_info.append(field_info)

        return cls(fields_info, sub_configs)



import typing
from betterconf.provider import AbstractProvider, DEFAULT_PROVIDER
from betterconf._config import ConfigInner, ConfigProto, Prefix

class_T = typing.TypeVar("class_T", bound=type)


@typing.overload
def betterconf(
    cls: class_T,
    provider: typing.Optional[AbstractProvider] = None,
    prefix: typing.Optional[typing.Union[Prefix, str]] = None,
    subconfig: bool = False,
) -> class_T: ...


@typing.overload
def betterconf(
    cls: None = None,
    provider: typing.Optional[AbstractProvider] = None,
    prefix: typing.Optional[typing.Union[Prefix, str]] = None,
    subconfig: bool = False,
) -> typing.Callable[[class_T], class_T]: ...


def betterconf(
    cls: typing.Optional[class_T] = None,
    provider: typing.Optional[AbstractProvider] = None,
    prefix: typing.Optional[typing.Union[Prefix, str]] = None,
    subconfig: bool = False,
) -> typing.Union[class_T, typing.Callable[[class_T], class_T]]:
    def inner(cls: class_T) -> class_T:
        def __init__(
            self: ConfigProto,
            _provider_: typing.Optional[AbstractProvider] = None,
            **to_override: typing.Any,
        ):
            for field in self.__bc_inner__.fields:
                if not field.field.provider:
                    field.field.provider = provider

                if _provider_:
                    field.field.provider = _provider_

                if field.name_in_python in to_override:
                    field.field.default = to_override[field.name_in_python]
                    setattr(self, field.name_in_python, field.field.value)
                else:
                    setattr(self, field.name_in_python, field.field.value)

            for sub_config in self.__bc_inner__.sub_configs:
                if not sub_config.cfg.__bc_provider__:
                    sub_config.cfg.__bc_provider__ = provider

                if _provider_:
                    sub_config.cfg.__bc_provider__ = _provider_

                if sub_config.name in to_override:
                    setattr(self, sub_config.name, to_override[sub_config.name])
                else:
                    config = sub_config.cfg(**to_override)
                    setattr(self, sub_config.name, config)

        nonlocal provider
        if subconfig is False:
            provider = provider or DEFAULT_PROVIDER

        nonlocal prefix
        if isinstance(prefix, str):
            prefix = Prefix(prefix)

        cls.__bc_subconfig__ = subconfig
        cls.__bc_inner__ = ConfigInner.parse_into(cls, provider, prefix)
        cls.__bc_prefix__ = prefix
        cls.__bc_provider__ = provider

        setattr(cls, "__init__", __init__)
        return cls

    if cls is None:
        return inner

    return inner(cls)



import os
import json
import typing

from pathlib import Path
from betterconf.exceptions import BetterconfError, VariableNotFoundError


class AbstractProvider:
    """Implement this class and pass to `field`"""

    def get(self, name: str) -> str:
        """Return a value (str) or raise a `VariableNotFoundError`"""
        raise NotImplementedError()


class EnvironmentProvider(AbstractProvider):
    """Default provider. Gets vals from environment"""

    def get(self, name: str) -> str:
        value = os.getenv(name)
        if value is None:
            raise VariableNotFoundError(name)
        return value


class JSONProvider(AbstractProvider):
    @staticmethod
    def __bool_object_hook(inp: typing.Any) -> typing.Dict[typing.Any, typing.Any]:
        d: typing.Dict[typing.Any, typing.Any] = {}
        for k, v in inp.items():
            if isinstance(v, bool):
                d[k] = str(v)
                continue
            elif isinstance(v, list):
                d[k] = json.dumps(v)
                continue
            d[k] = v
        return d

    def __init__(self, inp: str, nested_access: str = "."):
        # dirty hack cause betterconf itself deserializes objects and we have to implement clear interface based on
        # str`s
        self._content: typing.Union[typing.Any, typing.Dict[str, typing.Any]] = (
            json.loads(
                inp,
                object_hook=self.__bool_object_hook,
                parse_int=lambda i: i,
                parse_float=lambda f: f,
                parse_constant=lambda c: c,
            )
        )
        self._nested_access = nested_access
        if not isinstance(self._content, dict):
            raise ValueError("JSONProvider doesn't know how to operate not on dicts")

    @classmethod
    def from_path(cls, path: str, nested_access: str = ".") -> typing.Self:
        return cls.from_file(open(path, mode="r"), nested_access)

    @classmethod
    def from_file(cls, file: typing.IO[str], nested_access: str = ".") -> typing.Self:
        contents = file.read()
        file.close()
        return cls(contents, nested_access)

    @classmethod
    def from_string(cls, inp: str, nested_access: str = "."):
        return cls(inp, nested_access)

    def get(self, name: str) -> str:
        nested = name.split(self._nested_access)
        result: typing.Optional[str] = None
        storage = self._content

        "hello.world == {'hello': {'world': 123}'"
        for k in nested:
            result = storage.get(k)  # type: ignore
            storage = result  # type: ignore

        if result is None or not isinstance(result, str):
            raise VariableNotFoundError(name)

        return result


class DotenvProvider(AbstractProvider):
    def __init__(
        self, file_path: str | Path = ".env", *, auto_load: bool = False
    ) -> None:
        self.file_path = file_path
        self._loaded_into: typing.Literal["env", "in", None] = None

        self._environ = EnvironmentProvider()
        self._inner: dict[str, str] = {}

        self._auto_load = auto_load

    def _put_lines_to_vars(self, into: typing.Literal["env", "in"]):
        with open(self.file_path, "r") as f:
            lines = f.readlines()

        vars: dict[str, str] = {}
        for line in lines:
            if line == "\n":
                continue
            try:
                var_name, var_value = line.split("=", 1)
            except ValueError:
                raise BetterconfError(
                    "DotenvProvider can't read your dotenv file because it seems to be broken"
                )

            var_value = var_value.rstrip("\n")
            var_name = var_name.rstrip("\n")
            vars[var_name] = var_value

        self._loaded_into = into
        if into == "in":
            self._inner.update(vars)
        elif into == "env":
            os.environ.update(vars)

    def load_into_env(self):
        self._put_lines_to_vars(into="env")

    def load_into_provider(self):
        self._put_lines_to_vars(into="in")

    def get(self, name: str) -> str:
        if self._auto_load and not self._loaded_into:
            self.load_into_provider()

        if not self._loaded_into:
            raise BetterconfError("You haven't loaded values from .env manually")

        if self._loaded_into == "in":
            value = self._inner.get(name)
            if value is None:
                raise VariableNotFoundError(name)
            return value
        else:
            return self._environ.get(name)


DEFAULT_PROVIDER = EnvironmentProvider()



"""
Python configs made smooth and easy.
"""

from .decorator import betterconf
from ._config import Prefix
from ._field import field, Field, constant_field, reference_field, value
from ._specials import Alias
from .provider import (
    AbstractProvider,
    JSONProvider,
    EnvironmentProvider,
    DotenvProvider,
)
from .caster import (
    to_int,
    to_bool,
    to_list,
    to_float,
    to_loguru_log_level,
    to_logging_log_level,
    AbstractCaster,
)
from .exceptions import (
    ImpossibleToCastError,
    BetterconfError,
    VariableNotFoundError,
)

__author__ = "prostomarkeloff"
__all__ = (
    "betterconf",
    "field",
    "Field",
    "constant_field",
    "value",
    "reference_field",
    "AbstractProvider",
    "JSONProvider",
    "Alias",
    "Prefix",
    "EnvironmentProvider",
    "to_int",
    "to_bool",
    "to_list",
    "to_float",
    "to_loguru_log_level",
    "to_logging_log_level",
    "AbstractCaster",
    "BetterconfError",
    "VariableNotFoundError",
    "ImpossibleToCastError",
    "DotenvProvider",
    "__author__",
)



import typing
import logging

from betterconf.exceptions import ImpossibleToCastError

VT = typing.TypeVar("VT")


class AbstractCaster:
    def cast(self, val: str) -> typing.Union[typing.Any, typing.NoReturn]:
        """Try to cast or return val"""
        raise NotImplementedError()


class ConstantCaster(AbstractCaster, typing.Generic[VT]):
    ABLE_TO_CAST: typing.Dict[
        typing.Union[str, typing.Tuple[str, ...]], typing.Any
    ] = {}

    def cast(self, val: str) -> typing.Union[VT, typing.NoReturn]:
        """Cast using ABLE_TO_CAST dictionary as in BoolCaster"""
        if val in self.ABLE_TO_CAST:
            converted = self.ABLE_TO_CAST.get(val.lower())
            converted = typing.cast(VT, converted)
            return converted
        else:
            for key in self.ABLE_TO_CAST:
                if isinstance(key, tuple) and val.lower() in key:
                    return self.ABLE_TO_CAST[key]
                elif isinstance(key, str) and val.lower() == key:
                    return self.ABLE_TO_CAST[key]
            raise ImpossibleToCastError(val, self)


class BoolCaster(ConstantCaster[bool]):
    ABLE_TO_CAST = {
        "true": True,
        "1": True,
        "yes": True,
        "ok": True,
        "on": True,
        "false": False,
        "0": False,
        "no": False,
        "off": False,
    }


class IntCaster(AbstractCaster):
    def cast(self, val: str) -> typing.Union[int, typing.NoReturn]:
        try:
            as_int = int(val)
            return as_int
        except ValueError:
            raise ImpossibleToCastError(val, self)


class FloatCaster(AbstractCaster):
    def cast(self, val: str) -> typing.Union[float, typing.NoReturn]:
        val = val.replace(",", ".")
        try:
            as_float = float(val)
            return as_float
        except ValueError:
            raise ImpossibleToCastError(val, self)


class ListCaster(AbstractCaster):
    def __init__(self, separator: str = ","):
        self.separator = separator

    def cast(self, val: str) -> typing.List[typing.Any]:
        if val.endswith(self.separator):
            val = val[0 : len(val) - len(self.separator)]
        return val.split(self.separator)


class LoggingLogLevelCaster(ConstantCaster[int]):
    ABLE_TO_CAST = {
        "CRITICAL": logging.CRITICAL,
        "FATAL": logging.FATAL,
        "ERROR": logging.ERROR,
        "WARN": logging.WARN,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "NOTSET": logging.NOTSET,
    }


class LoguruLogLevelCaster(ConstantCaster[str]):
    levels = ["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]
    ABLE_TO_CAST = {level.lower(): level for level in levels}


class NothingCaster(AbstractCaster):
    """Caster who does nothing"""

    def cast(self, val: str) -> str:
        return val


to_bool = BoolCaster()
to_int = IntCaster()
to_float = FloatCaster()
to_list = ListCaster()
to_logging_log_level = LoggingLogLevelCaster()
to_loguru_log_level = LoguruLogLevelCaster()
DEFAULT_CASTER = NothingCaster()

BUILTIN_CASTERS = {
    bool: to_bool,
    int: to_int,
    float: to_float,
    list: to_list,
    str: DEFAULT_CASTER,  # for type hints
}

__all__ = (
    "to_bool",
    "to_int",
    "to_float",
    "to_list",
    "to_logging_log_level",
    "to_loguru_log_level",
    "AbstractCaster",
    "ConstantCaster",
    "DEFAULT_CASTER",
)



import typing
from typing import TypeVarTuple

from betterconf.caster import AbstractCaster
from betterconf.caster import DEFAULT_CASTER
from betterconf.exceptions import VariableNotFoundError, ImpossibleToCastError
from betterconf.provider import DEFAULT_PROVIDER, AbstractProvider


class Sentinel:
    @classmethod
    def is_setinel(cls, obj: typing.Any) -> bool:
        return isinstance(obj, cls)


_NO_DEFAULT = Sentinel()

T = typing.TypeVar("T")
Ts = TypeVarTuple("Ts")
SentinelOrT = typing.Union[Sentinel, T]


class _Field(typing.Generic[T]):
    # NB: fields are:
    # 1. lazy evaluated (when .value is called, I'm sure like always when initializing), fields are
    # 2. one-timers: every time the .value is called the recompute their value (that's needed for inheritance or
    # initializing this config with new params)

    def __init__(
        self,
        name: typing.Optional[str] = None,
        default: SentinelOrT[T] = _NO_DEFAULT,
        provider: typing.Optional[AbstractProvider] = None,
        caster: AbstractCaster = DEFAULT_CASTER,
        ignore_caster_error: bool = False,
    ):
        self.name = name
        self.provider = provider
        self.default = default
        self.caster = caster
        self.ignore_caster_error = ignore_caster_error

    def _get_value(self) -> T:
        try:
            if self.name is None:
                raise VariableNotFoundError("No name was given, as is a default value")

            if not self.provider:
                self.provider = DEFAULT_PROVIDER

            inner_value = self.provider.get(self.name)
        except VariableNotFoundError as e:
            if isinstance(self.default, Sentinel):
                raise e

            if callable(self.default):
                return self.default()
            else:
                return self.default
        try:
            casted = self.caster.cast(inner_value)

        except ImpossibleToCastError as e:
            if self.ignore_caster_error:
                return e.val  # type: ignore

            else:
                raise e

        else:
            return casted

    @property
    def value(self) -> T:
        return self._get_value()

    # can be used as `default=`
    def __call__(self, *_, **__: typing.Any):
        return self.value


if typing.TYPE_CHECKING:
    Field = typing.Annotated[T, ...]
else:
    Field = _Field


def value(
    name: typing.Optional[str] = None,
    default: T | typing.Any = _NO_DEFAULT,
    provider: typing.Optional[AbstractProvider] = None,
    caster: AbstractCaster = DEFAULT_CASTER,
    ignore_caster_error: bool = False,
) -> T:
    """
    Get a field and a value exactly when it's created
    """
    f = _Field[T](name, default, provider, caster, ignore_caster_error)
    return f.value


def field(
    name: typing.Optional[str] = None,
    default: T | typing.Any = _NO_DEFAULT,
    provider: typing.Optional[AbstractProvider] = None,
    caster: AbstractCaster = DEFAULT_CASTER,
    ignore_caster_error: bool = False,
) -> T:
    """
    Create a field for your config
    """
    return typing.cast(
        T, _Field[T](name, default, provider, caster, ignore_caster_error)
    )


def reference_field(*fields: *Ts, func: typing.Callable[[*Ts], T]) -> T:
    def _default() -> T:
        vars: typing.List[typing.Any] = []
        for field in fields:
            if isinstance(field, _Field):
                vars.append(field.value)  # type: ignore
            else:
                vars.append(field)

        return func(*vars)  # type: ignore

    return typing.cast(T, _Field(default=_default))


def constant_field(const: T) -> T:
    return typing.cast(T, _Field(default=const))


__all__ = ("Field", "field", "value", "reference_field", "constant_field")
